/*
 This file is part of the BlueJ program. 
 Copyright (C) 1999-2009,2010,2011  Michael Kolling and John Rosenberg 

 This program is free software; you can redistribute it and/or 
 modify it under the terms of the GNU General Public License 
 as published by the Free Software Foundation; either version 2 
 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 GNU General Public License for more details. 

 You should have received a copy of the GNU General Public License 
 along with this program; if not, write to the Free Software 
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA. 

 This file is subject to the Classpath exception as provided in the  
 LICENSE.txt file that accompanied this code.
 */
package bluej.compiler;

import java.io.*;
import java.util.regex.*;
import java.util.ArrayList;
import java.net.URI;
import javax.tools.Diagnostic;
import javax.tools.JavaFileObject;

import bluej.Config;
import bluej.utility.Debug;
import bluej.parser.ErrorParser;

/**
 * Code to check compiler errors/warnings and Java source code
 * and convert into more useful Diagnostic objects.
 * 
 * @author Warren Toomey
 */

public class ErrorCheck
{
    /*
     * errorList is a list of known errors, their type
     * and any regular expression required to detect
     * them from the error message generated by the
     * compiler. Shared between ErrorCheck objects.
     */
    private static ArrayList<String []> errorList = new ArrayList<String []>();

    /*
     * foundError is only set true if we find an error, either in the
     * source code or from a compiler message. We need this because
     * the DiagnosticListener inner class in the CompilerAPICompiler
     * class can't touch the result variable in the compile() method.
     */
    private boolean foundError= false;

    public boolean didHaveError()
    {
      return(foundError);
    }


    /**
     * Read in the list of known errors from the configuration file
     * errorlist.defs.
     */
    private void readErrorList()
    {
	File errorListFile = new File(Config.getBlueJLibDir(),
			Config.language + File.separator + "errorlist.defs");
        BufferedReader in;

	// Open the file up
        try {
		in = new BufferedReader( new FileReader( errorListFile ));

		// Read the lines in the file. Split each line on the ::
		// separator, and save the resulting array into the errorList.
		// Skip those input lines which begin with Java comments, and
		// also those which do not have exactly 3 fields
		while (in.ready()) {
			String line = in.readLine();
			if (line != null && line.startsWith("//")) continue;
			String [] fields = line.split("::");
			if (fields.length!=3) {
				Debug.message("Malformed line in " +
					"errorlist.defs" + ": " + line);
				continue;
			}
			errorList.add(fields);
		}

		// All done, close the file down
		in.close();
        }
        catch(Exception e) {
                Debug.message("ErrorCheck unable to load error list file " +
						"errorlist.defs: " + e);
        }
    }



    /**
     * Given a list of source code files, check them for errors
     * by parsing each one with an internal parser. Return
     * true if an error was found, false otherwise.
     * If an error was found, propagate a BlueJ Diagnostic object
     * about the error to the observer given as a parameter.
     */
    public boolean sourcesCheck(File[] sources, CompileObserver observer)
    {
      bluej.compiler.Diagnostic bjDiag;

      // Parse each of the source files. If we recognise any parse
      // error, we can send a Diagnostic object to the observer.
      for (int i=0; i < sources.length; i++) {
	ErrorParser eParse= null;
	// Try to open the file with the parser
	try { eParse = new ErrorParser(sources[i]); }
        catch (Exception e) { }
	if (eParse == null) continue;

	// Parse the file. Continue if no error found
	bjDiag= eParse.parse();
	if (bjDiag==null) continue;
	Debug.message("After parse, " + bjDiag);

	// If we can find a known error, send it to the observer
	if (findError("parse", bjDiag.getMessage(), bjDiag) != null) {
	  Debug.message("Known parse error, " + bjDiag);
	  observer.compilerMessage(bjDiag);
	}
      }

      return(foundError);
    }


    /**
     * Given a source code file and a compiler Diagnostic object which
     * represents an error or warning found by the compiler, interpret
     * the error or warning and propagate a BlueJ Diagnostic object
     * to the given observer.
     */
    public void compileCheck(File f,
	Diagnostic<? extends JavaFileObject> diag, CompileObserver observer)
    {
	bluej.compiler.Diagnostic bjDiag;
        String src = null;
        String message = diag.getMessage(null);
        if (diag.getSource() == null)
	  return;

	// With JDK 6, diag.getSource().getName()  apparently just returns the base
	// name without a path. To get the path we need to ask for the URI.
	//     However:
	// With JDK 7, the diag.getSource().toURI() returns an unusable URI if the
	// path is a UNC path (\\server\sharename\projdir\somefile.java).

	if (Config.isJava17()) {
	    src = diag.getSource().getName();
	}
	else {
	    // See bug: http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6419926
	    // JDK6 returns URIs without a scheme in some cases, so always resolve against a
	    // known "file:/" URI:
	    URI srcUri = f.toURI().resolve(diag.getSource().toUri());
	    src = new File(srcUri).getPath();
	}

	// Convert the Diagnostic object into a BlueJ Diagnostic object
	long startCol= diag.getColumnNumber();
	long endCol= diag.getEndPosition() - diag.getPosition() + startCol;
	long startLine= diag.getLineNumber();
	long endLine= diag.getLineNumber();
	int errtype = (diag.getKind() == Diagnostic.Kind.WARNING) ?
		bluej.compiler.Diagnostic.WARNING :
		bluej.compiler.Diagnostic.ERROR;
	bjDiag= new bluej.compiler.Diagnostic(errtype, null, src,   // Dummies
				startLine, startCol, endLine, endCol, null);

	// If we can find a known error, send it to the observer
        if (findError("comp", message, bjDiag)!= null) {
	  Debug.message("Compiler gives " + bjDiag);
          observer.compilerMessage(bjDiag);
	}
    }


    /**
     * Given a String which contains either a compiler error/warning message
     * or aJava source file, search it for known errors. Return a
     * bluej.compiler.Diagnostic reference on an error or warning, or
     * null if none found.
     * @param type	Type of error: "parse", "code" or "comp"
     * @param msg	The message from the compiler, or a Java source file
     * @param diag	Diagnostic from the compiler, or null if not
     */
    private bluej.compiler.Diagnostic findError(String type, String msg,
			bluej.compiler.Diagnostic bjDiag)
    {
	// Read in the error list if it is empty
	if (errorList.isEmpty())
	    readErrorList();
Debug.message("Got msg " + msg);

	// Search for each known error in the list, but
	// only apply the right ones, i.e. "code" ones
	// when we have src code, "comp" ones for compiler errors
	for (int i=0; i < errorList.size(); i++) {
	    String [] fields = (String [])errorList.get(i);
	    if (!fields[0].startsWith(type)) continue;

	    // Now apply the regex field to the message and see if it matches
	    Pattern p = Pattern.compile(fields[2], Pattern.DOTALL | Pattern.MULTILINE);
	    Matcher m = p.matcher(msg);
	    if (m.find()) {
	        ArrayList<String> context= new ArrayList<String>();
		String title= fields[1];
		String src= null;
		long startLine= 0, endLine= 0, startCol= 0, endCol= 0;
		int errtype;

		// Set any context words which were found by groups in the
		// regexp, ignoring group(0).
		for (int g = 0; g < m.groupCount(); g++) {
		    context.add(m.group(g+1));
		}

		if (fields[0].endsWith("err")) {
		  foundError= true;
		  errtype= bluej.compiler.Diagnostic.ERROR;
		} else {
		  errtype= bluej.compiler.Diagnostic.WARNING;
		}

		// If we don't have a BlueJ Diagnostic, build one
		// Calculate filename and the line and column numbers
		if (bjDiag == null) {
		    // Using the start and end positions of the match,
		    // count newlines to deduce line and column numbers.
		    int startPosn = m.start();
		    int endPosn = m.end() - 1;
		    int line=1, col=1;
		    for (int posn=0; posn <= endPosn; posn++) {
		        if (msg.charAt(posn)=='\n') {
			    line++; col=1;
			}
			if (posn==startPosn) {
			    startCol= col; startLine= line;
			}
			col++;
		    }
		    endCol= col; endLine= line;

		    bjDiag= new bluej.compiler.Diagnostic(errtype, title,
			null, startLine, startCol, endLine, endCol,
			context);
		} else {
		    // Modify the BlueJ Diagnostic which we received
		    bjDiag.setType(errtype);
		    bjDiag.setMessage(title);
		    bjDiag.setContext(context);
		}

		return(bjDiag);
	    }
	}

	// We didn't recognise the input error. If it's a compiler
	// error, return the raw message as we have to return someting.
	// Otherwise, just ignore it.
	if (type.equals("comp")) {
	  bjDiag.setMessage(msg);
	  return(bjDiag);
	} else
          return(null);
    }

}
