/*
 This file is part of the BlueJ program.
 Copyright (C) 1999-2010,2011  Michael Kolling and John Rosenberg

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

 This file is subject to the Classpath exception as provided in the
 LICENSE.txt file that accompanied this code.
 */
package bluej.parser;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.lang.Boolean;
import java.util.ArrayDeque;
import java.util.Iterator;

import bluej.parser.lexer.JavaTokenTypes;
import bluej.parser.lexer.LocatableToken;
import static bluej.parser.JavaErrorCodes.*;

import bluej.utility.Debug;

/**
 * A minimal parser which simply detects parse errors,
 * and is used during compilation to find mistakes
 * that can be shown to the programmer.
 *
 * @author Warren Toomey
 */
public class ErrorParser extends JavaParser {
    // For each ErrorParser we keep the File being parsed, and
    // only one parse error for the parse.
    private File file;
    private bluej.compiler.Diagnostic bjDiag;

    // It would be nice to simply record when we are in
    // an If condition, but anonymous classes muck this
    // up. We push TRUE or FALSE on this stack as is required to
    // indicate when we are in an If condition.
    private ArrayDeque<Boolean> ifCondStack = new ArrayDeque<Boolean>();


    // Keep track of the For loop variables in each of the nested For
    // loops that we have. We find and store the loop variables in
    // wantForVar as we parse the For declaration, then push wantForVar
    // as we enter the loop body.
    private ArrayDeque<String> forVarStack = new ArrayDeque<String>();
    boolean wantForVar = false;
    String wantForId = null;



    // Create a BufferedReader on the file, and no error yet
    public ErrorParser(File f) throws FileNotFoundException {
        super(new BufferedReader(
                  new InputStreamReader(
                      new FileInputStream(f))));
        bjDiag = null;
        file = f;
    }

    // Parse the whole compilation unit and return
    // any error that we found
    public bluej.compiler.Diagnostic parse() {
        parseCU();
        return(bjDiag);
    }

    @Override
    protected void error(String msg, int beginLine, int beginColumn, int endLine, int endColumn) {
        // Don't try to record more than 1 parse error
        if (bjDiag != null) return;

        // This is the first parse error, so save it as a bluej.compiler.Diagnostic
        bjDiag= new bluej.compiler.Diagnostic(
            bluej.compiler.Diagnostic.ERROR,
            msg, file.getPath(),
            beginLine, beginColumn, endLine, endColumn, null);
        Debug.message("In error, got " + msg + ", line " + beginLine);
    }

    private void error(String msg, LocatableToken token) {
        error(msg, token.getLine(), token.getColumn(), token.getEndLine(), token.getEndColumn());
    }

    /********************* If Processing *****************************/

    // We are inside an If condition between calls to
    // beginIfStmt() and beginIfCondBlock()
    @Override
    protected void beginIfStmt(LocatableToken token) {
        ifCondStack.push(Boolean.TRUE);
        super.beginIfStmt(token);
    }

    @Override
    protected void beginIfCondBlock(LocatableToken token) {
        if (!ifCondStack.isEmpty())
            ifCondStack.pop();
        super.beginIfCondBlock(token);
    }

    @Override
    protected void beginAnonClassBody(LocatableToken token, boolean isEnumMember) {
        // No longer in an If condition
        ifCondStack.push(Boolean.FALSE);
        super.beginAnonClassBody(token, isEnumMember);
    }

    @Override
    protected void endAnonClassBody(LocatableToken token, boolean included) {
        // Return the If condition state to what it was before
        if (!ifCondStack.isEmpty())
            ifCondStack.pop();
        super.endAnonClassBody(token, included);
    }

    @Override
    protected void gotBinaryOperator(LocatableToken token) {
        LocatableToken token2, token3;

        switch (token.getType()) {
        case JavaTokenTypes.ASSIGN:
	    // => or =< are wrong way around!
            token2= tokenStream.LA(1);
            if (token2==null)
                break;
            switch (token2.getType()) {
            case JavaTokenTypes.GT:
            case JavaTokenTypes.LT:
                error(BJ081 + token2.getText(), token);
            }

            // Error if we are trying to do an assignment
            // inside an If condition
            if (ifCondStack.peek() == Boolean.TRUE) {
                // Error: Assignment in If Condition
                error(BJ079, token);
            }
            break;

        case JavaTokenTypes.BAND:
        case JavaTokenTypes.BOR:
            // Error if we use bitwise AND or OR in an If Condition
            if (ifCondStack.peek() == Boolean.TRUE) {
                // Error: Bitwise Operator in If Condition: & or |
                error(BJ080 + token.getText(), token);
            }
            break;

        case JavaTokenTypes.EQUAL:
        case JavaTokenTypes.NOT_EQUAL:
            // String Comparison processing
            // == or != followed by "literal" and not followed by DOT is bad
            token2= tokenStream.LA(1);
            token3= tokenStream.LA(2);
            if (token2 != null && token2.getType() == JavaTokenTypes.STRING_LITERAL
                    && token3 != null && token3.getType() != JavaTokenTypes.DOT) {
                // Debug.message("== or != followed by String literal");
                // Bad String Comparison
                error(BJ084, token);
            }
            break;

        case JavaTokenTypes.LOR:
        case JavaTokenTypes.LAND:
            // && or || followed by a comparison operator is bad
            token2= tokenStream.LA(1);
            if (token2==null)
                break;
            switch (token2.getType()) {
            case JavaTokenTypes.EQUAL:
            case JavaTokenTypes.NOT_EQUAL:
            case JavaTokenTypes.LT:
            case JavaTokenTypes.LE:
            case JavaTokenTypes.GT:
            case JavaTokenTypes.GE:
                // Bad Comparison After And or Or
                error(BJ085 + token.getText() + " " + token2.getText(), token);
            }
            break;
        }

        super.gotBinaryOperator(token);
    }


    /********************* For Processing *****************************/

    // Determine if a given token type is an assignment or
    // a change to the variable preceding it
    private boolean isAssignment(int ttype) {
        return( ttype == JavaTokenTypes.ASSIGN
                || ttype == JavaTokenTypes.BAND_ASSIGN
                || ttype == JavaTokenTypes.BOR_ASSIGN
                || ttype == JavaTokenTypes.BSR_ASSIGN
                || ttype == JavaTokenTypes.BXOR_ASSIGN
                || ttype == JavaTokenTypes.DIV_ASSIGN
                || ttype == JavaTokenTypes.MINUS_ASSIGN
                || ttype == JavaTokenTypes.MOD_ASSIGN
                || ttype == JavaTokenTypes.PLUS_ASSIGN
                || ttype == JavaTokenTypes.SL_ASSIGN
                || ttype == JavaTokenTypes.SR_ASSIGN
                || ttype == JavaTokenTypes.STAR_ASSIGN
                || ttype == JavaTokenTypes.INC
                || ttype == JavaTokenTypes.DEC);
    }

    // We want to record the loop variable
    @Override
    protected void beginForLoop(LocatableToken token) {
        wantForVar= true;
        wantForId= null;
        super.beginForLoop(token);
    }

    // Found one, so temporarily record it
    @Override
    protected void gotForInit(LocatableToken first, LocatableToken idToken) {
        if (wantForVar) {
            wantForVar= false;
            wantForId = idToken.getText();
        }
    }

    // We have to delay saving the loop variable name until
    // we hit the loop body, so we don't trigger on expressions
    // in the second and third section of the loop definition
    @Override
    protected void beginForLoopBody(LocatableToken token) {
        // Debug.message("Pushed this identifier: " + wantForId);

        // Java allows 'for(;;)', so the wantForId could be null,
        // so make up a bogus value for it.
        if (wantForId == null)
            wantForId= "";
        forVarStack.push(wantForId);
    }


    // Remove the loop variable that we previously pushed
    @Override
    protected void endForLoop(LocatableToken token, boolean included) {
        if (!forVarStack.isEmpty())
            forVarStack.pop();
        wantForVar= false;
        wantForId= null;
        super.endForLoop(token, included);
    }

    // We are looking for two types of expressions:
    // 1. Expressions which occur in the first section
    // of the 'for' declaration, as this will name the loop variable.
    // 2. Expressions in the loop body which may alter the loop variable.
    @Override
    public void parseExpression() {
        // Get the next two tokens
        LocatableToken token= tokenStream.LA(1);
        LocatableToken token2= tokenStream.LA(2);

        // If the first one is an identifier
        if (token != null && token.getType() == JavaTokenTypes.IDENT) {

            // Record the variable if we are waiting for it in the For parsing
            if (wantForVar) {
                wantForVar= false;
                wantForId= token.getText();
            }

            // Otherwise, it is a general variable. See if it matches
            // any of the For variables that we have recorded
            else if (token2 != null && isAssignment(token2.getType())) {
                String id = token.getText();
                // Debug.message("Need to check this variable against For vars: " + id);
                for (Iterator iter = forVarStack.iterator(); iter.hasNext(); ) {
                    if (id.equals( iter.next() )) {
                        // Debug.message("Error about: " + id);
                        error(BJ082 + ": " + id, token);
                    }
                }
            }
        }
        super.parseExpression();
    }


    /****************** Bad String Processing *************************/

    @Override
    protected void gotLiteral(LocatableToken token) {
        if (token.getType() == JavaTokenTypes.STRING_LITERAL) {
            LocatableToken token2= tokenStream.LA(1);
            if ((token2.getType() == JavaTokenTypes.EQUAL)
                    || (token2.getType() == JavaTokenTypes.NOT_EQUAL)) {
                // Debug.message("String literal followed by == or !=");
                // Bad String Comparison
                error(BJ084, token);
            }
        }
        super.gotLiteral(token);
    }
}
